#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <csignal>
#include <cmath>
#include <string>

#include <unistd.h>

#include "io.h"
#include "autoptr.h"

enum {
	kbup=-100,
	kbdn,
	kblt,
	kbrt,
	kbhome,
	kbend,
	kbpgup,
	kbpgdn,
	kbback,
	kbdel,
	kbtab
};

typedef long double evaltype;
typedef long long   longtype;

static void eraseblanks(string &s)
{
	unsigned a, b=s.size();
	for(a=0; a<b && s[a]==' '; ++a);
	s.erase(0, a);
}

enum
{
	prio_none = 0,
	prio_add,
	prio_modulo,
	prio_times,
	prio_nega,
	prio_expon
};

typedef autoptr<class expr> exprp;
class expr : public ptrable
{
public:
	bool ok;
	expr() : ok(true)
	{
	}
	exprp load(string &s, int base, int prio=0);
	virtual evaltype eval() const
	{
		return 0xDEADBEEF;
	}
};
class expr_const : public expr
{
	evaltype d;
public:
	expr_const(evaltype dd) : d(dd) { }
	virtual evaltype eval() const { return d; }
};
class expr_2op : public expr
{
protected:
	exprp ll, rr;
public:
	expr_2op(exprp l, exprp r) : ll(l), rr(r)
	{
		ok = ll && ll->ok && rr && rr->ok;
	}
	virtual ~expr_2op() { }
};
class expr_minus : public expr_2op
{
public:
	expr_minus(exprp ll, exprp rr) : expr_2op(ll, rr) { }
	virtual evaltype eval() const { return ll->eval() - rr->eval(); }
};
class expr_plus : public expr_2op
{
public:
	expr_plus(exprp ll, exprp rr) : expr_2op(ll, rr) { }
	virtual evaltype eval() const { return ll->eval() + rr->eval(); }
};
class expr_times : public expr_2op
{
public:
	expr_times(exprp ll, exprp rr) : expr_2op(ll, rr) { }
	virtual evaltype eval() const { return ll->eval() * rr->eval(); }
};
class expr_divid : public expr_2op
{
public:
	expr_divid(exprp ll, exprp rr) : expr_2op(ll, rr) { }
	virtual evaltype eval() const { return ll->eval() / rr->eval(); }
};
class expr_modul : public expr_2op
{
public:
	expr_modul(exprp ll, exprp rr) : expr_2op(ll, rr) { }
	virtual evaltype eval() const { return fmodl(ll->eval(), rr->eval()); }
};
class expr_expon : public expr_2op
{
public:
	expr_expon(exprp ll, exprp rr) : expr_2op(ll, rr) { }
	virtual evaltype eval() const { return powl(ll->eval(), rr->eval()); }
};
exprp expr::load(string &s, int base, int prio)
{
	eraseblanks(s);
	if(!s.size())return NULL;
	exprp left;
	
	if(s[0]=='(')
	{
		s.erase(0, 1);
		left = load(s, base, 0);
		eraseblanks(s);
		if(s[0] != ')')return NULL;
		s.erase(0, 1);
	}
	else if(s[0] == '-') // && prio < prio_nega -- this exception exists in qb also
	{
		s.erase(0, 1);
		left = new expr_minus(new expr_const(0), load(s, base, prio_nega));
	}
	else
	{
        evaltype d=0, kerta=1;
        bool piste=false, jotain=false;
		while(s.size())
        {
            unsigned n;
            char c = toupper(s[0]);
            if(c >= '0' && c <= '9')
            {
                if(base < 10 && c-'0' >= base)return NULL;
                n = c - '0';
            }
            else if(c >= 'A' && c <= 'Z')
            {
                if(base <= 10 || c-'A' >= base-10)return NULL;
                n = c + (10 - 'A');
            }
            else
            {
                if(c != '.' || piste)break;
                piste = true;
                s.erase(0, 1);
                continue;
            }
            s.erase(0, 1);
            jotain = true;
            if(piste)
                d += n / (kerta *= base);
            else
                d = d*base + n;
        }
	    left = jotain ? new expr_const(d) : NULL;
	}
reop:
	if(!left || !left->ok)return NULL;
	eraseblanks(s);
	if(!s.size())return left;
	if(s[0] == '-' && prio < prio_add) { s.erase(0, 1); left = new expr_minus(left, load(s, base, prio_add)); goto reop; }
	if(s[0] == '+' && prio < prio_add) { s.erase(0, 1); left = new expr_plus(left,  load(s, base, prio_add)); goto reop; }
	if(s[0] == '*' && prio < prio_times) { s.erase(0, 1); left = new expr_times(left,  load(s, base, prio_times)); goto reop; }
	if(s[0] == '/' && prio < prio_times) { s.erase(0, 1); left = new expr_divid(left,  load(s, base, prio_times)); goto reop; }
	if(s[0] == '%' && prio < prio_modulo) { s.erase(0, 1); left = new expr_modul(left,  load(s, base, prio_modulo)); goto reop; }
	if(s[0] == '^' && prio < prio_expon) { s.erase(0, 1); left = new expr_expon(left,  load(s, base, prio_expon)); goto reop; }
	return left;
}

class inputfield
{
	unsigned x, y, width, cpos;
public:
	unsigned base;
	string content;
	bool selected;
public:
	inputfield(unsigned ax, unsigned ay, unsigned awidth, unsigned abase) :
		x(ax), y(ay), width(awidth), base(abase), content("0"), selected(false)
	{
		key(kbend);
	}
	void draw(unsigned tick=0) const
	{
		unsigned a;
		char buf[5] = "(00:";
		
		if(base == 2) strncpy(buf, "BIN", 3);
		else if(base == 8) strncpy(buf, "OCT", 3);
		else if(base ==10) strncpy(buf, "DEC", 3);
		else if(base ==16) strncpy(buf, "HEX", 3);
		else if(base ==36) strcpy(buf, "MEGA");
		else 
		{
			buf[1] += base/10;
			buf[2] += base%10;
			buf[3] = ')';
		}
		
		SetChar(x+0, y, buf[0], 0x17,0,0);
		SetChar(x+1, y, buf[1], 0x17,0,0);
		SetChar(x+2, y, buf[2], 0x17,0,0);
		SetChar(x+3, y, buf[3], 0x17,0,0);
		
		for(a=0; a<width; ++a)
		{
			char c = a < content.size() ? content[a] : ' ';
			SetChar(x+a+5, y, c, selected ? tick<30 && a == cpos ? 0x71 : 0x0B : 0x03, 0,0);
		}
	}
	void key(int c)
	{
		switch(c)
		{
			case kblt:
				if(cpos > 0)cpos--;
				break;
			case kbrt:
				if(cpos < content.size())cpos++;
				break;
			case kbhome:
				cpos = 0;
				break;
			case kbend:
				cpos = content.size();
				break;
			case kbback:
				if(cpos > 0)cpos--;
				/* through */
			case kbdel:
				content.erase(cpos, 1);
				break;
			default:
				if(c >= 32)
				{
					content.insert(cpos, 1, (char)c);
					cpos++;
				}
		}
	}
public:
	void laske(string s, unsigned srcbase)
	{
		exprp tmp = new expr();
		tmp = tmp->load(s, srcbase, 0);
		if(tmp && !s.size())
		{
			static const char ss[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
			evaltype d = tmp->eval();
			bool nega = d < 0;
			
			if(nega)d = -d;
			
			longtype num = (longtype)d;
			d -= num;
			content = "";
			for(;;)
			{
				content = ss[num%base] + content;
				num /= base;
				if(!num)break;
			}
			if(nega) content = "-" + content;
			if(d > 0.0 && d < 1.0)
			{
				content += '.';
				for(unsigned i=0; i<width; ++i)
				{
					d *= base;
					num = (longtype)d;
					content += ss[num];
					d -= num;
					if(d == 0.0)break;
				}
			}
		}
		else
			content = "Error: '" + s + "'";
	}
};

class kaikki
{
	inputfield *inputs[20];
	unsigned sele;
	
public:
	kaikki(void) : sele(8)
	{
    	for(unsigned y=0; y<20; ++y)
    	{
    		unsigned base = y+2;
    		if(base > 10)base += 1;
    		if(base > 12)base += base / 7;
    		if(base > 16)base = base*2 - 14;
    		inputs[y] = new inputfield(2, 2 + y, 31, base);
    	}
	}
	
	void idlet()
	{
		unsigned tick=0;
		for(;;)
		{
			for(unsigned y=0; y<20; ++y)
			{
				inputs[y]->selected = y==sele;
				inputs[y]->draw(tick);
			}				
			if(kbhit())break;
			UpdateVideo(1);
			usleep(20000);
			if(++tick==40)tick=20;
		}
	}
	
	void laske(bool kaikki = false)
	{
		for(unsigned y=0; y<20; ++y)
			if(y != sele || kaikki)
			{
				inputs[y]->laske(inputs[sele]->content, inputs[sele]->base);
				inputs[y]->key(kbend);
			}
	}
	
	void key(int c)
	{
		if(c == kbup)
		{
			if(sele > 0)sele--;
		}
		else if(c == kbdn)
		{
			if(sele < 19)sele++;
		}
		else if(c == kbpgup)
			sele = 0;
		else if(c == kbpgdn)
			sele = 19;
		else if(c == '\n' || c == '\r')
			laske(true);
		else
		{
			inputs[sele]->key(c);
			laske();
		}
	}
} Kaikki;

static int getkey()
{
	static int buffer = -1;
	int ch;
	
	if(buffer >= 0)
	{
		ch = buffer;
		buffer = -1;
		return ch;
	}
	
	Kaikki.idlet();
	ch = getch();
	
	if(ch == '')return kbhome;
	if(ch == '')return kbend;
	if(ch == '')return kblt;
	if(ch == '')return kbrt;
	if(ch == '')return kbpgup;
	if(ch == '')return kbpgdn;
	if(ch == '')return kbback;
	if(ch == '')return kbback;
	if(ch == '')return kbdel;
	if(ch == 9)return kbtab;
	if(ch == '')return kbup;
	if(ch == '')return kbdn;
	if(ch == 27)
	{
		unsigned n = 0;
		Kaikki.idlet();
		ch = getch();
		if(ch != '[') { buffer = ch; return 27; }
		for(;;)
		{
			Kaikki.idlet();
			ch = getch();
			if(ch < '0' || ch > '9')break;
			n = n*10 + ch-'0';
		}
		if(ch == 'A')return kbup;
		if(ch == 'B')return kbdn;
		if(ch == 'C')return kbrt;
		if(ch == 'D')return kblt;
		if(ch == '~')
			switch(n)
			{
				case 1: return kbhome;
				case 3: return kbdel;
				case 4: return kbend;
				case 5: return kbpgup;
				case 6: return kbpgdn;
			}
		return 27;
	}
	return ch;
}

static void Quittah(int n)
{
    n=n;DoneVideo();
    exit(-1);
}

void Center(int xadd, int y, int c,int tr,int time, const char *fmt, ...)
{
	char Buf[512], *s;
	va_list ap;
	int nx, x;
	
	va_start(ap, fmt);	
	#ifdef HAVE_VSNPRINTF
	vsnprintf(Buf, sizeof(Buf)-1, fmt, ap);
	#else
	vsprintf(Buf, fmt, ap);
	#endif
	va_end(ap);
	
	s=Buf;
	
GetNx:
	for(x=0; s[x] && s[x]!='\n'; x++);
	x = nx = xadd + (41-xadd-x) / 2;
	
	for(; *s; s++)
		switch(*s)
		{
			case '\b':
				x--;
				break;
			case '\n':
				y++;
				s++;
				goto GetNx;
				break;
			case '\r':
				x=nx;
				break;
			default:
				SetChar(x++,y,*s, c,tr,time);
		}
}

int main(void)
{
	InstallDriver(&vt100, NULL);
	SelectDriver(DEFAULT_IO);

	InitVideo();
	
	signal(SIGINT, Quittah);
	
	for(unsigned y=0; y<24; ++y)
		for(unsigned x=0; x<40; ++x)
		{
			unsigned c=0;
			if(!y || y==23)
			{
				if(x > 0)c |= c_lt;
				if(x < 39)c |= c_rt;
			}
			if(!x || x==39)
			{
				if(y > 0)c |= c_up;
				if(y < 23)c |= c_dn;
			}
			SetChar(x,y, corner(c),
				c ? 0x3F : 0x11,
			    0,0);
		}
		
	Center(5, 4, 0x0A, 0, 100,
		"Multiple base calculator\n"
		"and value converter");
		
	Center(5, 18, 0x08, 0, 70, 
		"Copyright (C) 1992,2000\n"
		"Bisqwit");
	Center(5, 20, 0x08, 0, 500,
		"http://iki.fi/bisqwit/");

	// This is the help text.
	// It will leave after many seconds.
	Center(5, 13, 0x07, 0, 300,
	    "** HELP **\n"
		"Enter a numeric expression,\n"
		"and you see its value.\n"
		"Use arrows keys\n"
		"to move the cursor.");
	
	// This will mask the help text away
	// for the first some seconds.
	Center(5, 13, 0x0C, 0, 68,
	    "          \n"
		"                           \n"
		"                      \n"
		"               \n"
		"                   ");
		
	Center(0,22, 0x18, 0,1200,
		"NES-style output module. Yeah. :)");
	Center(0,22, 0x17, 0,200,
		"                                 ");
	
	Center(0,1, 0x19, 0,0,   "---- GPL. See the file COPYING. ----");
	Center(0,1, 0x17, 0,140, "                                    ");
	
	for(;;)
	{
		int c = getkey();
		
		if(c == 27)break;
		
		Kaikki.key(c);
	}
	
	DoneVideo();
	
	return 0;
}

